#include <stdio.h>
#include <stdlib.h>


const char strDelims[] = " \t\n.!?,;:\"";


const char strOrig[] =
"You don't care for me
I don't a-care about that
Got a new fool, ha
I'd like to laugh at
I have only one burning desire,
Let me stand next to your fire";

char *strTmp;
char *strParsed;
char *strRes;


typedef struct _tnode {
  char *word;
  int  count;
  struct _tnode *less;
  struct _tnode *greater;
} tnode;


tnode *add_tnode(tnode *, const char *);
int get_tnode_count(tnode *, const char *);
void free_tnodes(tnode *);


int main(void)
{
  char *ptr, *prev;
  int i, iOrigLen;
  tnode *troot;

  iOrigLen = sizeof(strOrig) - 1;

  strRes    = (char *) malloc(iOrigLen + 1);
  strParsed = (char *) malloc(iOrigLen + 1);
  strTmp    = (char *) malloc(iOrigLen + 1);
  if ( !strRes || !strParsed || !strTmp ) {
    fputs("Not enough memory!", stderr);
    return -1;
  }
  memcpy(strParsed, strOrig, iOrigLen + 1);
  memcpy(strTmp,    strOrig, iOrigLen + 1);

  /* Parse the text and arrange the words in a tree, counting number
     of occurences of each word */
  ptr = (char *) strtok(strParsed, strDelims);
  troot = NULL;
  while ( ptr ) {
    troot = add_tnode(troot, ptr);
    ptr = (char *) strtok(NULL, strDelims);
  }

  /* Parse the text again, this time checking with our tree whether
     the word occurs more than once; if yes, copy it into strRes; if
     not, leave it out */
  i = 0;
  prev = (char *) strOrig;
  ptr = (char *) strtok(strTmp, strDelims);

  while ( ptr ) { /* loop while we get any words */
    /* copy strOrig from where we left off erlier up to ptr */
    while ( (prev - strOrig) < (ptr - strTmp) )
      *(strRes + i++) = *prev++;
    /* skip the word if it occurs only once */
    if ( get_tnode_count(troot, ptr) < 2 )
      prev += strlen(ptr);
    /* parse out next word */
    ptr = (char *) strtok(NULL, strDelims);
  }

  /* copy the remaining part of strOrig */
  ptr = (char *) strOrig + iOrigLen;
    while ( (prev - strOrig) <= (ptr - strTmp) ) /* Include the terminating '\0' */
      *(strRes + i++) = *prev++;

  printf("\n***Original string:\n\n%s\n\n***Resulting string:\n\n%s\n\n",
         strOrig,
         strRes);

  /* Free the mallocs! */
  free_tnodes(troot);
  free(strTmp);
  free(strParsed);
  free(strRes);

  return 0;
}


/*
 * tnode *add_tnode(tnode *node, char *word)
 *   Adds new word to the tree, or increments word's count if it is
 *   already there.  If fails to allocate space for a new node, just
 *   bails out with exit code -1.  Note:  this is a recursive func.
 *
 * Takes:
 *   tnode *   pointer to node of the tree;
 *   char *    pointer to the word.
 *
 * Returns:
 *   pointer to the newly added node, or the pointer passed in if it
 *   nothing was allocated.
 */

tnode *add_tnode(tnode *node, const char *word)
{
  int cmp;

  if ( !node ) {
    node = (tnode *) malloc(sizeof(tnode));
    if ( !node )
      exit(-1);
    node->word = (char *) word;
    node->count = 1;
    node->less = NULL;
    node->greater = NULL;
  } else if ( ( cmp = strcmp(word, node->word) ) == 0 )
    node->count++;
  else if ( cmp < 0 )
    node->less = add_tnode(node->less, word);
  else
    node->greater = add_tnode(node->greater, word);

  return node;
}


/*
 * int get_tnode_count(tnode *node, char *word)
 *   Returns number of occurences of the word stored in the tree.
 *   Note:  this is a recursive function.
 *
 * Takes:
 *   tnode *   pointer to node;
 *   char *    word to be looked up.
 *
 * Returns:
 *   number of occurences of the word.
 */

int get_tnode_count(tnode *node, const char *word)
{
  int cmp;

  if ( !node ) /* no such word found in the tree! */
    return 0;

  if ( ( cmp = strcmp(word, node->word) ) == 0 )
    return node->count;
  else if ( cmp < 0 )
    return get_tnode_count(node->less, word);
  else
    return get_tnode_count(node->greater, word);
}


/*
 * void free_tnodes(tnode *node)
 *   Frees the whole tree rooted at the node.  Note:  this function is
 *   recursive.
 *
 * Takes:
 *   tnode *   pointer to the (root) node.
 *
 * Returns:
 *   nothin'
 */

void free_tnodes(tnode *node)
{
  if ( node ) {
    free_tnodes(node->less);
    free_tnodes(node->greater);
    free(node);
  }

  return;
}


/*********** Emacs vars ***********/
/* Local variables: */
/* compile-command: "gcc 5_16.c" */
/* End: */
